@@ -2,12 +2,12 @@
 #include "log.cpp"
 
 static MESSAGE_TYPE MsgStream=MSG_STDOUT;
+static RAR_CHARSET RedirectCharset=RCH_DEFAULT;
+static bool ProhibitInput=false;
+
 const int MaxMsgSize=2*NM+2048;
 
-#ifdef _WIN_ALL
 static bool StdoutRedirected=false,StderrRedirected=false,StdinRedirected=false;
-#endif
-
 
 #ifdef _WIN_ALL
 static bool IsRedirected(DWORD nStdHandle)
@@ -42,16 +42,32 @@ void InitConsole()
   if (!StderrRedirected)
     _setmode(_fileno(stderr), _O_U16TEXT);
 #endif
+#elif defined(_UNIX)
+  StdoutRedirected=!isatty(fileno(stdout));
+  StderrRedirected=!isatty(fileno(stderr));
+  StdinRedirected=!isatty(fileno(stdin));
 #endif
 }
 
 
-void InitConsoleOptions(MESSAGE_TYPE MsgStream)
+void SetConsoleMsgStream(MESSAGE_TYPE MsgStream)
 {
   ::MsgStream=MsgStream;
 }
 
 
+void SetConsoleRedirectCharset(RAR_CHARSET RedirectCharset)
+{
+  ::RedirectCharset=RedirectCharset;
+}
+
+
+void ProhibitConsoleInput()
+{
+  ProhibitInput=true;
+}
+
+
 #ifndef SILENT
 static void cvt_wprintf(FILE *dest,const wchar *fmt,va_list arglist)
 {
@@ -63,17 +79,26 @@ static void cvt_wprintf(FILE *dest,const wchar *fmt,va_list arglist)
   safebuf wchar Msg[MaxMsgSize];
   if (dest==stdout && StdoutRedirected || dest==stderr && StderrRedirected)
   {
-    // Avoid Unicode for redirect in Windows, it does not work with pipes.
-    vswprintf(Msg,ASIZE(Msg),fmtw,arglist);
-    safebuf char MsgA[MaxMsgSize];
-    WideToChar(Msg,MsgA,ASIZE(MsgA));
-    CharToOemA(MsgA,MsgA); // Console tools like 'more' expect OEM encoding.
-
-    // We already converted \n to \r\n above, so we use WriteFile instead
-    // of C library to avoid unnecessary additional conversion.
     HANDLE hOut=GetStdHandle(dest==stdout ? STD_OUTPUT_HANDLE:STD_ERROR_HANDLE);
+    vswprintf(Msg,ASIZE(Msg),fmtw,arglist);
     DWORD Written;
-    WriteFile(hOut,MsgA,(DWORD)strlen(MsgA),&Written,NULL);
+    if (RedirectCharset==RCH_UNICODE)
+      WriteFile(hOut,Msg,(DWORD)wcslen(Msg)*sizeof(*Msg),&Written,NULL);
+    else
+    {
+      // Avoid Unicode for redirect in Windows, it does not work with pipes.
+      safebuf char MsgA[MaxMsgSize];
+      if (RedirectCharset==RCH_UTF8)
+        WideToUtf(Msg,MsgA,ASIZE(MsgA));
+      else
+        WideToChar(Msg,MsgA,ASIZE(MsgA));
+      if (RedirectCharset==RCH_DEFAULT || RedirectCharset==RCH_OEM)
+        CharToOemA(MsgA,MsgA); // Console tools like 'more' expect OEM encoding.
+
+      // We already converted \n to \r\n above, so we use WriteFile instead
+      // of C library to avoid unnecessary additional conversion.
+      WriteFile(hOut,MsgA,(DWORD)strlen(MsgA),&Written,NULL);
+    }
     return;
   }
   // MSVC2008 vfwprintf writes every character to console separately
@@ -123,36 +148,54 @@ void eprintf(const wchar *fmt,...)
 
 
 #ifndef SILENT
+static void QuitIfInputProhibited()
+{
+  // We cannot handle user prompts if -si is used to read file or archive data
+  // from stdin.
+  if (ProhibitInput)
+  {
+    mprintf(St(MStdinNoInput));
+    ErrHandler.Exit(RARX_FATAL);
+  }
+}
+
+
 static void GetPasswordText(wchar *Str,uint MaxLength)
 {
   if (MaxLength==0)
     return;
+  QuitIfInputProhibited();
+  if (StdinRedirected)
+    getwstr(Str,MaxLength); // Read from pipe or redirected file.
+  else
+  {
 #ifdef _WIN_ALL
-  HANDLE hConIn=GetStdHandle(STD_INPUT_HANDLE);
-  HANDLE hConOut=GetStdHandle(STD_OUTPUT_HANDLE);
-  DWORD ConInMode,ConOutMode;
-  DWORD Read=0;
-  GetConsoleMode(hConIn,&ConInMode);
-  GetConsoleMode(hConOut,&ConOutMode);
-  SetConsoleMode(hConIn,ENABLE_LINE_INPUT);
-  SetConsoleMode(hConOut,ENABLE_PROCESSED_OUTPUT|ENABLE_WRAP_AT_EOL_OUTPUT);
+    HANDLE hConIn=GetStdHandle(STD_INPUT_HANDLE);
+    HANDLE hConOut=GetStdHandle(STD_OUTPUT_HANDLE);
+    DWORD ConInMode,ConOutMode;
+    DWORD Read=0;
+    GetConsoleMode(hConIn,&ConInMode);
+    GetConsoleMode(hConOut,&ConOutMode);
+    SetConsoleMode(hConIn,ENABLE_LINE_INPUT);
+    SetConsoleMode(hConOut,ENABLE_PROCESSED_OUTPUT|ENABLE_WRAP_AT_EOL_OUTPUT);
 
-  ReadConsole(hConIn,Str,MaxLength-1,&Read,NULL);
-  Str[Read]=0;
-  SetConsoleMode(hConIn,ConInMode);
-  SetConsoleMode(hConOut,ConOutMode);
+    ReadConsole(hConIn,Str,MaxLength-1,&Read,NULL);
+    Str[Read]=0;
+    SetConsoleMode(hConIn,ConInMode);
+    SetConsoleMode(hConOut,ConOutMode);
 #else
-  char StrA[MAXPASSWORD];
+    char StrA[MAXPASSWORD*4]; // "*4" for multibyte UTF-8 characters.
 #if defined(_EMX) || defined (__VMS)
-  fgets(StrA,ASIZE(StrA)-1,stdin);
+    fgets(StrA,ASIZE(StrA)-1,stdin);
 #elif defined(__sun)
-  strncpyz(StrA,getpassphrase(""),ASIZE(StrA));
+    strncpyz(StrA,getpassphrase(""),ASIZE(StrA));
 #else
-  strncpyz(StrA,getpass(""),ASIZE(StrA));
+    strncpyz(StrA,getpass(""),ASIZE(StrA));
 #endif
-  CharToWide(StrA,Str,MaxLength);
-  cleandata(StrA,sizeof(StrA));
+    CharToWide(StrA,Str,MaxLength);
+    cleandata(StrA,sizeof(StrA));
 #endif
+  }
   Str[MaxLength-1]=0;
   RemoveLF(Str);
 }
@@ -162,20 +205,27 @@ static void GetPasswordText(wchar *Str,uint MaxLength)
 #ifndef SILENT
 bool GetConsolePassword(UIPASSWORD_TYPE Type,const wchar *FileName,SecPassword *Password)
 {
-  uiAlarm(UIALARM_QUESTION);
+  if (!StdinRedirected)
+    uiAlarm(UIALARM_QUESTION);
   
   while (true)
   {
-    if (Type==UIPASSWORD_GLOBAL)
-      eprintf(L"\n%s: ",St(MAskPsw));
-    else
-      eprintf(St(MAskPswFor),FileName);
+//    if (!StdinRedirected)
+      if (Type==UIPASSWORD_GLOBAL)
+        eprintf(L"\n%s: ",St(MAskPsw));
+      else
+        eprintf(St(MAskPswFor),FileName);
 
-    wchar PlainPsw[MAXPASSWORD];
+    wchar PlainPsw[MAXPASSWORD+1];
     GetPasswordText(PlainPsw,ASIZE(PlainPsw));
     if (*PlainPsw==0 && Type==UIPASSWORD_GLOBAL)
       return false;
-    if (Type==UIPASSWORD_GLOBAL)
+    if (wcslen(PlainPsw)>=MAXPASSWORD)
+    {
+      PlainPsw[MAXPASSWORD-1]=0;
+      uiMsg(UIERROR_TRUNCPSW,MAXPASSWORD-1);
+    }
+    if (!StdinRedirected && Type==UIPASSWORD_GLOBAL)
     {
       eprintf(St(MReAskPsw));
       wchar CmpStr[MAXPASSWORD];
@@ -204,6 +254,8 @@ bool getwstr(wchar *str,size_t n)
   // Print buffered prompt title function before waiting for input.
   fflush(stderr);
 
+  QuitIfInputProhibited();
+
   *str=0;
 #if defined(_WIN_ALL)
   // fgetws does not work well with non-English text in Windows,
@@ -215,6 +267,7 @@ bool getwstr(wchar *str,size_t n)
     Array<char> StrA(n*4); // Up to 4 UTF-8 characters per wchar_t.
     File SrcFile;
     SrcFile.SetHandleType(FILE_HANDLESTD);
+    SrcFile.SetLineInputMode(true);
     int ReadSize=SrcFile.Read(&StrA[0],StrA.Size()-1);
     if (ReadSize<=0)
     {
@@ -222,8 +275,15 @@ bool getwstr(wchar *str,size_t n)
       // calling Ask(), so let's better exit.
       ErrHandler.Exit(RARX_USERBREAK);
     }
-    StrA[ReadSize-1]=0;
+    StrA[ReadSize]=0;
+
+    // We expect ANSI encoding here, but "echo text|rar ..." to pipe to RAR,
+    // such as send passwords, we get OEM encoding by default, unless we
+    // use "chcp" in console. But we avoid OEM to ANSI conversion,
+    // because we also want to handle ANSI files redirection correctly,
+    // like "rar ... < ansifile.txt".
     CharToWide(&StrA[0],str,n);
+    cleandata(&StrA[0],StrA.Size()); // We can use this function to enter passwords.
   }
   else
   {
@@ -279,7 +339,7 @@ int Ask(const wchar *AskStr)
 
   for (int I=0;I<NumItems;I++)
   {
-    eprintf(I==0 ? (NumItems>4 ? L"\n":L" "):L", ");
+    eprintf(I==0 ? (NumItems>3 ? L"\n":L" "):L", ");
     int KeyPos=ItemKeyPos[I];
     for (int J=0;J<KeyPos;J++)
       eprintf(L"%c",Item[I][J]);
